import { Component, EventEmitter, Output, Input, forwardRef, ElementRef, Renderer2, ViewChild, HostListener, isDevMode, OnChanges, HostBinding, ChangeDetectorRef, ChangeDetectionStrategy } from '@angular/core';
import { DateHelper } from './utilities/date-helper';
import { I18nService } from './services/i18n.service';
import { L10N_PREFIX, L10nService } from './services/l10n.service';
import { noop, approximateStringMatching, minValidator, maxValidator, isDocumentAvailable, isValidRange, guid } from './utilities/utils';
import { cloneDate, isInRange, isEqual, dateInRange, setTime } from './utilities/date-utils';
import { dateFormatRegExp, timePartRegExp } from './utilities/regular-expressions';
import { CbxInputWrapperComponent } from '../cbx-input-wrapper/cbx-input-wrapper.component';




var getTime = function (date) { return date ? date.getTime() : null; };
var firstChange = function (change) { return change ? change.firstChange : true; };



var Arrow;
(function (Arrow) {
    Arrow[Arrow["Up"] = 0] = "Up";
    Arrow[Arrow["Down"] = 1] = "Down";
    Arrow[Arrow["None"] = 2] = "None";
})(Arrow || (Arrow = {}));

var Keys;
(function (Keys) {
    Keys[Keys["backspace"] = 8] = "backspace";
    Keys[Keys["tab"] = 9] = "tab";
    Keys[Keys["enter"] = 13] = "enter";
    Keys[Keys["shift"] = 16] = "shift";
    Keys[Keys["ctrl"] = 17] = "ctrl";
    Keys[Keys["alt"] = 18] = "alt";
    Keys[Keys["space"] = 32] = "space";
    Keys[Keys["page up"] = 33] = "page up";
    Keys[Keys["page down"] = 34] = "page down";
    Keys[Keys["end"] = 35] = "end";
    Keys[Keys["home"] = 36] = "home";
    Keys[Keys["left"] = 37] = "left";
    Keys[Keys["up"] = 38] = "up";
    Keys[Keys["right"] = 39] = "right";
    Keys[Keys["down"] = 40] = "down";
    Keys[Keys["delete"] = 46] = "delete";
})(Keys || (Keys = {}));

@Component({
    selector: 'cbx-dateinput',
    templateUrl: './cbx-input-date.component.html',
    styleUrls: ['./cbx-input-date.component.less'],

    changeDetection: ChangeDetectionStrategy.OnPush,
    exportAs: 'cbx-dateinput',
    providers: [
        { provide: L10N_PREFIX, useValue: 'cbx.dateinput' },
        L10nService
    ],

})
export class CbxInputDate implements OnChanges  {

    @Input('disabled') disabled: boolean;
    @Input() required: boolean = false;
    @Input('readonly') readonly: boolean;
    @Input('title') title: string;
    @Input('label') label: string;
    @Input()range:boolean;
    @Input() timePickerOnly: boolean;
    @ViewChild('inputWrapper') inputWrapper: CbxInputWrapperComponent;

    

    _tabIndex:number
    @Input('tabindex')
    get tabindex() {
        return this._tabIndex;
    }
    set tabindex(tabindex) {
        this._tabIndex = tabindex;
    }   

    @Input('format') format: string;
    @Input('formatPlaceholder') formatPlaceholder: string;
    @Input('placeholder') placeholder: string;

    _steps: any = {}
    @Input('steps')
    get steps() {
        return this._steps;
    }
    set steps(steps) {
        this._steps = steps || {};
    }

    _max:Date
    @Input('max')
    get max() {
        return this._max;
    }
    set max(max) {
        this._max = max;
        this.ariaValueMax = getTime(max);
    }
    _min: Date
    @Input('min')
    get min () {
        return this._min;
    }
    set min (min) {
        this._min = min;
        this.ariaValueMin = getTime(min);
    }


    @Input('rangeValidation') rangeValidation: boolean;
    @Input('autoCorrect') autoCorrect: boolean;

    _value:Date
    @Input('value')
    get value() {
        return this._value;
    }
    set value(value) {
        if (value && typeof value === 'string') {
            value = new Date(value);
        }
        
        this.verifyValue(value);
        if (this.autoCorrect && !isInRange(value, this.min, this.max)) {
            return;
        }
        this._value = cloneDate(value);
        this._value = value;
    }




    @Input('focusableId') focusableId: string;
    @Input('spinners') spinners: boolean;
    @Input('isPopupOpen') isPopupOpen: boolean;
    @Input('hasPopup') hasPopup: boolean;

    @Output() valueChange: EventEmitter<Date> = new EventEmitter();
    @Output() valueUpdate: EventEmitter<Date> = new EventEmitter();

    @Output() onFocus: EventEmitter<boolean> = new EventEmitter();
    @Output() onBlur: EventEmitter<boolean> = new EventEmitter();

    @ViewChild('dateInput') dateInput: ElementRef;

    @HostBinding('class.k-state-disabled') disabledClass: string;

    @HostBinding('class.k-widget')
    @HostBinding('class.k-dateinput')
    wrapperClass: string;

    @HostListener('mousewheel', ['$event'])
    @HostListener('onmousewheel', ['$event'])
    @HostListener('DOMMouseScroll', ['$event'])
    mousescroll(event) {
        this.handleMouseWheel(event);
    }

    @HostListener('keydown', ['$event']) 
    keydown(event) {
        if (this.disabled || this.readonly || event.altKey || event.ctrlKey || event.metaKey) {
            return;
        }
        if (event.keyCode === Keys.tab) {
            var moved = this.switchDateSegment(event.shiftKey ? -1 : 1);
            if (moved) {
                event.preventDefault();
            }
            return;
        }
        if (event.keyCode === Keys.backspace) {
            this.backspace = true;
            return;
        }
        if(event.key == "a"){
            this.toggleAMPM("AM");
            return;
        }
        if(event.key == "p"){
            this.toggleAMPM("PM");
            return;
        }

        switch (event.keyCode) {
            case Keys.down:
                this.modifyDateSegmentValue(-1);
                break;
            case Keys.up:
                this.modifyDateSegmentValue(1);
                break;
            case Keys.right:
                this.switchDateSegment(1);
                break;
            case Keys.left:
                this.switchDateSegment(-1);
                break;
            case Keys.home:
                this.selectNearestSegment(0);
                break;
            case Keys.end:
                this.selectNearestSegment(this.inputValue.length);
                break;
            default:
                return; //skip the preventDefault if we didn't handled the keyCode
        }
        event.preventDefault();
    }

    @HostListener('input') handleInput() {
        if (this.disabled || this.readonly) {
            return;
        }
        if (this.paste) {
            this.updateOnPaste();
            this.paste = false;
            return;
        }
        var diff = approximateStringMatching(this.currentValue, this.currentFormat, this.inputValue, this.caret()[0]);
        var navigationOnly = (diff.length === 1 && diff[0][1] === "_");
        var switchPart = false;
        if (!navigationOnly) {
            var parsedPart = void 0;
            for (var i = 0; i < diff.length; i++) {
                parsedPart = this.dateHelper.parsePart(diff[i][0], diff[i][1], this.resetSegmentValue);
                switchPart = parsedPart.switchToNext;
            }
            var candidate = this.dateHelper.getDateObject();
            if (this.value && candidate && !this.formatSections.date) {
                this.dateHelper = this.getDateHelper(setTime(this.value, candidate));
            }
        }
        this.resetSegmentValue = false;
        this.putDateInRange();
        this.updateElementValue(this.isActive);
        this.triggerChange();
        if (diff.length && diff[0][0] !== "_") {
            this.selectDateSegment(diff[0][0]);
        }
        if (switchPart || navigationOnly) {
            this.switchDateSegment(1);
        }
        if (this.backspace) {
            this.switchDateSegment(-1);
        }
        this.backspace = false;
    }

    @HostListener('paste') handlePaste() {
        this.paste = true;
    }
    

    arrow:any;
    arrowDirection:any;
    formatSections: any;
    isActive: boolean;
    hasMousedown: boolean;
    currentValue: string;
    currentFormat: string;
    backspace: boolean;
    resetSegmentValue: boolean;
    resolvedPromise: any;
    minValidateFn: any;
    maxValidateFn: any;
    date: null;
    paste: boolean;
    ngChange:any;
    ngTouched:any;
    symbolsMap:any;
    ariaValueMax: Date;
    ariaValueMin: Date;
    dateHelper: DateHelper;
    ariaValueNow: Date;
    ariaValueText: Date;
    intlSubscription: any;

    constructor(private changeDetectorRef: ChangeDetectorRef, private i18n: I18nService, private renderer: Renderer2, private element: ElementRef, private localization: L10nService ) {
        this.focusableId = "k-" + guid();
        this.disabled = false;
        this.readonly = false;
        this.title = "";
        this.tabindex = 0;
        this.format = "d";
        this.placeholder = "";
        this.rangeValidation = true;
        this.autoCorrect = false;
        this.spinners = false;
        this.isPopupOpen = false;
        this.hasPopup = false;
        this.valueChange = new EventEmitter();
        this.valueUpdate = new EventEmitter();
        this.onFocus = new EventEmitter(); 
        this.onBlur = new EventEmitter(); 
        
        this.arrow = Arrow;
        this.arrowDirection = Arrow.None;


        this.formatSections = { date: false, time: false };
        this.isActive = false;
        this.hasMousedown = false;


        this.currentValue = "";
        this.currentFormat = "";
        this.backspace = false;
        this.resetSegmentValue = true;
        this.resolvedPromise = Promise.resolve(null);
        this.minValidateFn = noop;
        this.maxValidateFn = noop;
        this._value = null;
        this.date = null;
        this.paste = false;
        this.ngChange = function (_) { };
        this.ngTouched = function () { };
        this.symbolsMap = this.dateSymbolMap();
        this.updateFormatSections();
    }

    get inputElement() {
        return this.dateInput ? this.dateInput.nativeElement : null;
    }

    get inputValue() {
        return (this.inputElement || {}).value || '';
    }

    isEmpty () {
        return false;
    }

    toggleAMPM(toggleVal:"AM"|"PM"){
        const inputElement:HTMLInputElement = this.inputElement;
        const inputVal:string = this.inputValue;
        const AMindex = inputVal.indexOf("AM");
        const PMindex = inputVal.indexOf("PM");

        //console.log("AM",inputVal.indexOf("AM"));
        //console.log("PM",inputVal.indexOf("PM"));
        //console.log("selection",inputElement.selectionStart);
        
        if(inputElement.selectionStart>=inputVal.indexOf("AM") && AMindex>-1 && toggleVal=="PM"){
            this.dateHelper.modifyPart("a", 1);
        }else if(inputElement.selectionStart>=inputVal.indexOf("PM") && PMindex>-1 && toggleVal=="AM"){
            this.dateHelper.modifyPart("a", -1);
        }
        
    }

    ngOnChanges(changes) {
        var _this = this;
        this.verifyRange();
        if (this.hasChanged(changes, 'min') || this.hasChanged(changes, 'max')) {
            this.minValidateFn = this.rangeValidation ? minValidator(this.min) : noop;
            this.maxValidateFn = this.rangeValidation ? maxValidator(this.max) : noop;
            if (!changes.value && (!firstChange(changes.min) || !firstChange(changes.max))) {
                this.resolvedPromise.then(function () { return _this.ngChange(cloneDate(_this.value)); });
            }
        }
        if (changes.format) {
            this.symbolsMap = this.dateSymbolMap();
            this.updateFormatSections();
        }
        var calcDate = this.dateHelper && isEqual(this.value, this.dateHelper.getDateObject());
        if (changes.format || !calcDate || changes.placeholder) {
            this.dateHelper = this.getDateHelper(this.value);
            this.updateElementValue(this.isActive);
        }
    }
    ngOnDestroy() {
        if (this.intlSubscription) {
            this.intlSubscription.unsubscribe();
        }
    }
    validate(control) {
        return this.minValidateFn(control) || this.maxValidateFn(control);
    }
    
    ngOnInit() {
        this.renderer.removeAttribute(this.element.nativeElement, 'tabindex');
        this.dateHelper = this.getDateHelper(this.value);
        this.updateElementValue(null);
        this.intlSubscription = this.i18n.changes.subscribe(this.intlChange.bind(this));
    }
    setDisabledState(isDisabled) {
        this.disabled = isDisabled;
        this.changeDetectorRef.markForCheck();
    }
    writeValue(value) {
        this.verifyValue(value);
        this.dateHelper = this.getDateHelper(value);
        this.value = cloneDate(value);
        this.valueUpdate.emit(cloneDate(value)); 
        this.updateElementValue(this.isActive);
    }
    
    triggerChange() {
        var value = this.dateHelper.getDateObject();
        if (+value !== +this.value) {
            this.value = cloneDate(value);
            this.notify();
        }
    }
    notify() {
        this.ngChange(cloneDate(this.value));
        this.valueChange.emit(cloneDate(this.value));
        this.valueUpdate.emit(cloneDate(this.value)); 
        
    }
    
    registerOnChange = function (fn) {
        this.ngChange = fn;
    }
    
    registerOnTouched = function (fn) {
        this.ngTouched = fn;
    }
    
    focus() {
        var input = this.inputElement;
        if (input) {
            input.focus();
            this.selectDateSegment(this.currentFormat[0]);
        }
    }
    blur() {
        var input = this.inputElement;
        if (input) {
            input.blur();
        }
    }
    handleButtonClick(offset) {
        this.arrowDirection = Arrow.None;
        this.modifyDateSegmentValue(offset);
    }
    
    modifyDateSegmentValue(offset) {
        var caret = this.caret();
        var symbol = this.currentFormat[caret[0]];
        var step = (this.steps || {})[this.symbolsMap[symbol]] || 1;
        this.dateHelper.modifyPart(symbol, offset * step);
        this.putDateInRange();
        this.updateElementValue(this.isActive);
        this.triggerChange();
        this.selectDateSegment(symbol);
    }
    
    switchDateSegment(offset) {
        var caret = this.caret();
        if (this.dateHelper.resetLeadingZero()) {
            this.updateElementValue(this.isActive);
        }
        if (caret[0] < caret[1] && this.currentFormat[caret[0]] !== this.currentFormat[caret[1] - 1]) {
            this.selectNearestSegment(offset > 0 ? caret[0] : caret[1] - 1);
            this.resetSegmentValue = true;
            return true;
        }
        var previousFormatSymbol = this.currentFormat[caret[0]];
        var a = caret[0] + offset;
        while (a > 0 && a < this.currentFormat.length) {
            if (this.currentFormat[a] !== previousFormatSymbol &&
                this.currentFormat[a] !== "_") {
                break;
            }
            a += offset;
        }
        if (this.currentFormat[a] === "_") {
            //there is not known symbol found
            return false;
        }
        var b = a;
        while (b >= 0 && b < this.currentFormat.length) {
            if (this.currentFormat[b] !== this.currentFormat[a]) {
                break;
            }
            b += offset;
        }
        if (a > b && (b + 1 !== caret[0] || a + 1 !== caret[1])) {
            this.caret(b + 1, a + 1);
            this.resetSegmentValue = true;
            return true;
        }
        else if (a < b && (a !== caret[0] || b !== caret[1])) {
            this.caret(a, b);
            this.resetSegmentValue = true;
            return true;
        }
        return false;
    }
    selectDateSegment(symbol) {
        var begin = -1;
        var end = 0;
        for (var i = 0; i < this.currentFormat.length; i++) {
            if (this.currentFormat[i] === symbol) {
                end = i + 1;
                if (begin === -1) {
                    begin = i;
                }
            }
        }
        if (begin < 0) {
            begin = 0;
        }
        this.caret(0, 0);
        this.caret(begin, end);
    }
    handleClick = function () {
        this.hasMousedown = false;
        if (this.isActive) {
            this.selectNearestSegment(this.caret()[0]);
        }
    }
    handleFocus = function () {
        this.isActive = true;
        this.updateElementValue();
        if (!this.hasMousedown) {
            this.caret(0, this.inputValue.length);
        }
        this.hasMousedown = false;
        this.onFocus.emit();
    }
    handleBlur = function () {
        this.isActive = false;
        this.resetSegmentValue = true;
        if (this.dateHelper.month && this.dateHelper.date ) {
            if (!this.dateHelper.year) {
                this.dateHelper.setExisting('y', false);
                this.dateHelper.year = true;
            }

            if (!this.timePickerOnly) {
                if (!this.dateHelper.hours) {
                    this.dateHelper.value.setHours(0);
                    this.dateHelper.setExisting('h', true);
                }
                if (!this.dateHelper.minutes) {
                    this.dateHelper.value.setMinutes(0);
                    this.dateHelper.setExisting('m', true);
                }

                if (!this.dateHelper.minutes) {
                    this.dateHelper.value.setMinutes(0);
                    this.dateHelper.setExisting('m', true);
                }
            }
        }
        this.dateHelper.resetLeadingZero();
        this.updateElementValue();
        this.ngTouched();
        this.onBlur.emit();
    }

    handleMouseWheel = function (event) {
        if (this.disabled || this.readonly || !this.isActive) {
            return;
        }
        event = window.event || event;
        if (event.shiftKey) {
            this.switchDateSegment((event.wheelDelta || -event.detail) > 0 ? -1 : 1);
        }
        else {
            this.modifyDateSegmentValue((event.wheelDelta || -event.detail) > 0 ? 1 : -1);
        }
        event.returnValue = false;
        if (event.preventDefault) {
            event.preventDefault();
        }
        if (event.stopPropagation) {
            event.stopPropagation();
        }
    }
    
    getDateHelper(value) {
        var leadingZero = (<any>((this.dateHelper || {}) || null)).leadingZero;

        var dateHelper = new DateHelper(this.i18n, this.formatPlaceholder, this.format, value);
        dateHelper.setLeadingZero(this.isActive ? leadingZero : null);
        return dateHelper;
    }
    
    dateSymbolMap() {
        var reducer = function (map, part) {
            map[part.pattern[0]] = part.type;
            return map;
        };
        return this.i18n.splitDateFormat(this.format, null).reduce(reducer, {});
    }
    
    updateElementValue(isActive) {
        var start = this.caret()[0]; 
        var texts = this.dateHelper.getTextAndFormat();
        var showPlaceholder = !this.isActive && this.placeholder && !this.dateHelper.hasValue();
        var input = this.inputElement;
        this.currentFormat = texts[1];
        this.currentValue = !showPlaceholder ? texts[0] : '';
        this.renderer.setProperty(input, "value", this.currentValue);
        if (input.placeholder !== this.placeholder) {
            this.renderer.setProperty(input, "placeholder", this.placeholder);
        }
        var currentDate = this.dateHelper.getDateObject();
        this.ariaValueNow = getTime(currentDate);
        this.ariaValueText = this.i18n.formatDate(currentDate, this.format, null);
        if (isActive) {
            this.selectNearestSegment(start);
        }
    }
    
    hasChanged(changes, field) {
        var change = changes[field];
        return change && !isEqual(change.previousValue, change.currentValue);
    }
    
    caret(start?: number, end?: number) {
        if (end === void 0) { end = start; }
        var isPosition = start !== undefined;
        var returnValue = [start, start];
        var element = this.inputElement;
        if (isPosition && (this.disabled || this.readonly)) {
            return undefined;
        }
        try {
            if (element.selectionStart !== undefined) {
                if (isPosition) {
                    if (isDocumentAvailable() && document.activeElement !== element) {
                        element.focus();
                    }
                    element.setSelectionRange(start, end);
                }
                returnValue = [element.selectionStart, element.selectionEnd];
            }
        }
        catch (e) {
            returnValue = [];
        }
        return returnValue;
    }
    selectNearestSegment(index) {
        // Finds the nearest (in both directions) known part.
        for (var i = index, j = index - 1; i < this.currentFormat.length || j >= 0; i++ , j--) {
            if (i < this.currentFormat.length && this.currentFormat[i] !== "_") {
                this.selectDateSegment(this.currentFormat[i]);
                return;
            }
            if (j >= 0 && this.currentFormat[j] !== "_") {
                this.selectDateSegment(this.currentFormat[j]);
                return;
            }
        }
    }
    
    verifyRange() {
        if (!isDevMode()) {
            return;
        }
        if (!isValidRange(this.min, this.max)) {
            throw new Error("The max value should be bigger than the min.");
        }
    }
    
    verifyValue(value) {
        if (!isDevMode()) {
            return;
        }
        if (value && !(value instanceof Date)) {
            throw new Error("The 'value' should be a valid JavaScript Date instance.");
        }
    }
    putDateInRange() {
        var currentDate = this.dateHelper.getDateObject();
        var candidate = dateInRange(currentDate, this.min, this.max);
        if (this.autoCorrect && !isEqual(currentDate, candidate)) {
            this.dateHelper = this.getDateHelper(candidate);
        }
    }
    updateFormatSections () {
        this.formatSections = this.i18n.splitDateFormat(this.format, null)
        .reduce(function (_a, p) {
            var date = _a.date, time = _a.time;
            return {
                date: date || dateFormatRegExp.test(p.type),
                time: time || timePartRegExp.test(p.type)
            };
        }, { date: false, time: false });
    }
    intlChange() {
        this.updateFormatSections();
        this.dateHelper = this.getDateHelper(this.value);
        this.updateElementValue(this.isActive);
    }
    updateOnPaste() {
        var value = this.i18n.parseDate(this.inputValue, this.format, null) || this.value;
        var notify = +value !== +this.value;
        this.writeValue(value);
        if (notify) {
            this.notify();
        }
    }

}
