import { Injectable, EventEmitter } from '@angular/core';
import defaultLocale from '../utilities/default-locale'; 
import { dateFormatRegExp, formatRegExp, timeSpecifiersRegExp, dateRegExp, offsetRegExp } from '../utilities/regular-expressions';
import { cloneLower } from '../utilities/date-utils';
import formatters from '../utilities/date-formatters';
import parsers from '../utilities/date-parsers';


var VALUE_FORMAT_LENGTH = {
    numeric: 1,
    "2-digit": 2,
    short: 3,
    long: 4,
    narrow: 5
};
var REMOVAL_PENALTY = 120;
var ADDITION_PENALTY = 20;
var LENGHT_DELTA = [2, 1, 5, 3, 4];
var PENALTIES = {};

@Injectable()
export class I18nService {
    EMPTY = "";
    DEFAULT_LOCALE = "en";
    LITERAL = 'literal';
    DATE_FIELD_MAP = {
        'G': 'era',
        'y': 'year',
        'q': 'quarter',
        'Q': 'quarter',
        'M': 'month',
        'L': 'month',
        'd': 'day',
        'E': 'weekday',
        'c': 'weekday',
        'e': 'weekday',
        'h': 'hour',
        'H': 'hour',
        'k': 'hour',
        'K': 'hour',
        'm': 'minute',
        's': 'second',
        'a': 'dayperiod',
        'x': 'zone',
        'X': 'zone',
        'z': 'zone',
        'Z': 'zone'
    };

    NAME_TYPES = {
        month: {
            type: 'months',
            minLength: 3,
            standAlone: 'L'
        },

        quarter: {
            type: 'quarters',
            minLength: 3,
            standAlone: 'q'
        },

        weekday: {
            type: 'days',
            minLength: {
                E: 0,
                c: 3,
                e: 3
            },
            standAlone: 'c'
        },

        dayperiod: {
            type: 'dayPeriods',
            minLength: 0
        },

        era: {
            type: 'eras',
            minLength: 0
        }
    };

    DATE_OPTIONS_MAP = [{
        key: "era",
        specifier: "G"
    }, {
        key: "year",
        specifier: "y"
    }, {
        key: "month",
        specifier: "M"
    }, {
        key: "day",
        specifier: "d"
    }, {
        key: "weekday",
        specifier: "E"
    }, {
        key: "hour",
        getSpecifier: function(options) {
            return options.hour12 ? "h" : "H";
        }
    }, {
        key: "minute",
        specifier: "m"
    }, {
        key: "second",
        specifier: "s"
    }, {
        key: "timeZoneName",
        specifier: "z"
    }];

    STAND_ALONE_SPECIFIERS = {
        e: 'c',
        E: 'c',
        M: 'L',
        Q: 'q'
    };

    FORMATS_SEQUENCE = ["G", "g", "F", "Y", "y", "M", "m", "D", "d", "y", "T", "t"];

    STANDARD_DATE_FORMATS = [
        "yyyy/MM/dd HH:mm:ss",
        "yyyy/MM/dd HH:mm",
        "yyyy/MM/dd",
        "E MMM dd yyyy HH:mm:ss",
        "yyyy-MM-ddTHH:mm:ss.SSSSSSSXXX",
        "yyyy-MM-ddTHH:mm:ss.SSSXXX",
        "yyyy-MM-ddTHH:mm:ss.SSXXX",
        "yyyy-MM-ddTHH:mm:ssXXX",
        "yyyy-MM-ddTHH:mm:ss.SSSSSSS",
        "yyyy-MM-ddTHH:mm:ss.SSS",
        "yyyy-MM-ddTHH:mmXXX",
        "yyyy-MM-ddTHH:mmX",
        "yyyy-MM-ddTHH:mm:ss",
        "yyyy-MM-ddTHH:mm",
        "yyyy-MM-dd HH:mm:ss",
        "yyyy-MM-dd HH:mm",
        "yyyy-MM-dd",
        "HH:mm:ss",
        "HH:mm"
    ];


    resolvedFormats = {};
    specifiersRegex = {};

    changes:EventEmitter<boolean>;

    constructor() {
        this.changes = new EventEmitter();
    }

    splitDateFormat(format, locale) {
        if (locale === void 0) locale = this.DEFAULT_LOCALE;

        var info = this.localeInfo(locale);
        var pattern = this.datePattern(format, info);
        var parts = [];
        var lastIndex = dateFormatRegExp.lastIndex = 0;
        var match = dateFormatRegExp.exec(pattern);

        while (match) {
            var value = match[0];

            if (lastIndex < match.index) {
                this.addLiteral(parts, pattern.substring(lastIndex, match.index));
            }

            if (value.startsWith('"') || value.startsWith("'")) {
                this.addLiteral(parts, value);
            } else {
                var specifier = value[0];
                var type = this.DATE_FIELD_MAP[specifier];
                var part = {
                    type: type,
                    pattern: value,
                    hour12: null,
                    names: null
                };

                if (type === 'hour') {
                    part.hour12 = this.isHour12(value);
                }

                var names = this.NAME_TYPES[type];

                if (names) {
                    var minLength = this.isNumber(names.minLength) ? names.minLength : names.minLength[specifier];
                    var patternLength = value.length;

                    if (patternLength >= minLength) {
                        part.names = {
                            type: names.type,
                            nameType: this.dateNameType(patternLength),
                            standAlone: names.standAlone === specifier
                        };
                    }
                }

                parts.push(part);
            }

            lastIndex = dateFormatRegExp.lastIndex;
            match = dateFormatRegExp.exec(pattern);
        }

        if (lastIndex < pattern.length) {
            this.addLiteral(parts, pattern.substring(lastIndex));
        }

        return parts;
    }

    localeInfo(locale) {
        return this.availableLocaleInfo("en-Latn-US", ["US"]);
    }

    
    availableLocaleInfo(fullName, suffixes) {
        var parts = fullName.split("-");
        var language = parts[0];
        var script = parts[1];
        var territory = parts[2];
        
        return defaultLocale[fullName] || (suffixes.indexOf(territory) !== -1 && defaultLocale[language + "-" + territory]) || (suffixes.indexOf(script) !== -1 && defaultLocale[language + "-" + script]) || defaultLocale[language];
    }
    datePattern(format, info) {
        var calendar = info.calendar;
        var result;
        if (this.isString(format)) {
            if (calendar.patterns[format]) {
                result = calendar.patterns[format];
            } else {
                result = format;
            }
        } else if (format) {
            if (format.pattern) {
                return format.pattern;
            }

            var skeleton = format.skeleton;
            if (!skeleton) {
                if (format.datetime) {
                    result = this.formatString(calendar.dateTimeFormats[format.datetime], calendar.timeFormats[format.datetime], calendar.dateFormats[format.datetime]);
                } else if (format.date) {
                    result = calendar.dateFormats[format.date];
                } else if (format.time) {
                    result = calendar.timeFormats[format.time];
                } else {
                    skeleton = this.skeletonFromOptions(format);
                }
            }

            if (skeleton) {
                result = this.skeletonFormat(skeleton, info);
            }
        }

        if (!result) {
            result = calendar.patterns.d;
        }

        return result;
    }

    addLiteral(parts, value) {
        var lastPart = parts[parts.length - 1];
        if (lastPart && lastPart.type === this.LITERAL) {
            lastPart.pattern += value;
        } else {
            parts.push({
                type: this.LITERAL,
                pattern: value
            });
        }
    }
    isHour12(pattern) {
        return pattern === 'h' || pattern === 'K';
    }

    isNumber(value) {
        return typeof value === "number";
    }
    isString(value) {
        return typeof value === "string";
    }
    isDate(value) {
        return Boolean(value) && typeof (value.getTime) === 'function' && typeof (value.getMonth) === 'function';
    }
    dateNameType(formatLength) {
        var nameType;
        if (formatLength <= 3) {
            nameType = "abbreviated";
        } else if (formatLength === 4) {
            nameType = "wide";
        } else if (formatLength === 5) {
            nameType = "narrow";
        } else if (formatLength === 6) {
            nameType = "short";
        }
        return nameType;
    }

    formatString(format, ...args: any[]) {
        var values = arguments;

        return format.replace(formatRegExp, function (match, index) {
            var value = values[parseInt(index, 10) + 1];

            return value;
        });
    }

    skeletonFromOptions(options) {
        var result = [];
        for (var idx = 0; idx < this.DATE_OPTIONS_MAP.length; idx++) {
            var option = this.DATE_OPTIONS_MAP[idx];
            var field = option.key;
            var value = options[field];
            if (value) {
                var spcifier = option.specifier || option.getSpecifier(options);
                result.push(spcifier.repeat(VALUE_FORMAT_LENGTH[value]));
            }
        }

        return result.join(this.EMPTY);
    }
    skeletonFormat(skeleton, info) {
        var availableFormats = info.calendar.dateTimeFormats.availableFormats;
        if (availableFormats[skeleton]) {
            return availableFormats[skeleton];
        }
        if (this.resolvedFormats[info.name] && this.resolvedFormats[info.name][skeleton]) {
            return this.resolvedFormats[info.name][skeleton];
        }
        var timeStartIndex = skeleton.search(timeSpecifiersRegExp);
        var result;
        if (timeStartIndex > 0) {
            var dateSkeleton = skeleton.substr(0, timeStartIndex);
            var timeSkeleton = skeleton.substr(timeStartIndex);

            result = this.formatString(info.calendar.dateTimeFormats.short, //should be deterimed based on specifiers
                availableFormats[timeSkeleton] || this.findBestMatch(this.skeletonSpecifiers(timeSkeleton), availableFormats),
                availableFormats[dateSkeleton] || this.findBestMatch(this.skeletonSpecifiers(dateSkeleton), availableFormats));
        } else {
            result = this.findBestMatch(this.skeletonSpecifiers(skeleton), availableFormats);
        }

        this.cacheFormat(skeleton, result, info.name);
        return result;
    }

    
    findBestMatch(specifiers, availableFormats) {
        var specifiersLength = specifiers.length;
        var maxScore = -Number.MAX_VALUE;
        var bestMatches, result;
        for (var format in availableFormats) {
            var matches = [];
            var currentFormat = format.replace("v", "z");
            var score = 0;
            for (var idx = 0; idx < specifiersLength; idx++) {
                var specifier = specifiers[idx];
                var specifierRegex = this.getSpecifierRegex(specifier[0]);
                var match = (specifierRegex.exec(currentFormat) || [])[0];

                if (!match) {
                    score -= REMOVAL_PENALTY;
                } else {
                    currentFormat = currentFormat.replace(match, this.EMPTY);
                    if (match.length !== specifier.length) {
                        var delta = Math.max(Math.min(LENGHT_DELTA[match.length] - LENGHT_DELTA[specifier.length], 2), -2);
                        score -= PENALTIES[delta];
                    }
                }

                matches.push(match);

                if (score < maxScore) {
                    break;
                }
            }

            if (currentFormat.length) {
                score -= this.skeletonSpecifiers(currentFormat).length * ADDITION_PENALTY;
            }

            if (score > maxScore) {
                maxScore = score;
                bestMatches = matches;
                result = availableFormats[format];
            }
        }

        result = result.replace("v", "z");

        for (var idx$1 = 0; idx$1 < specifiersLength; idx$1++) {
            var bestMatch = bestMatches[idx$1];
            if (bestMatch && bestMatch !== specifiers[idx$1]) {
                var matchSpecifier = bestMatches[idx$1][0];
                result = result.replace(this.getSpecifierRegex(matchSpecifier), specifiers[idx$1]);
                if (this.STAND_ALONE_SPECIFIERS[matchSpecifier]) {
                    result = result.replace(this.getSpecifierRegex(this.STAND_ALONE_SPECIFIERS[matchSpecifier]), specifiers[idx$1]);
                }
            }
        }

        return result;
    }
    skeletonSpecifiers(skeleton) {
        var result = [];
        var current = skeleton.charAt(0);
        var specifier = current;
        for (var idx = 1; idx < skeleton.length; idx++) {
            var character = skeleton.charAt(idx);
            if (character === specifier) {
                current += character;
            } else {
                result.push(current);
                current = specifier = character;
            }
        }

        result.push(current);

        return result;
    }

    getSpecifierRegex(specifier) {
        if (!this.specifiersRegex[specifier]) {
            this.specifiersRegex[specifier] = new RegExp(specifier + "+");
        }
        return this.specifiersRegex[specifier];
    }
    cacheFormat(skeleton, format, locale) {
        if (!this.resolvedFormats[locale]) {
            this.resolvedFormats[locale] = {};
        }
        this.resolvedFormats[locale][skeleton] = format;
    }

    formatDate(date, format, locale) {
        if (locale === void 0) locale = "en";

        if (!this.isDate(date)) {
            if (date === undefined || date === null) {
                return "";
            }
            return date;
        }

        var info = this.localeInfo(locale);
        var pattern = this.datePattern(format, info);

        return pattern.replace(dateFormatRegExp, function (match) {
            var formatLength = match.length;
            var result;

            if (match.includes("'") || match.includes("\"")) {
                result = match.slice(1, formatLength - 1);
            } else {
                result = formatters[match[0]](date, formatLength, info);
            }

            return result;
        });
    }
    parseDate(value, formats, locale) {
        if (locale === void 0) locale = this.DEFAULT_LOCALE;

        if (!value) {
            return null;
        }

        if (this.isDate(value)) {
            return value;
        }

        var parseValue = String(value).trim();
        var date = this.parseMicrosoftDateFormat(parseValue);
        if (date) {
            return date;
        }

        var info: any = this.localeInfo(locale);
        var parseFormats = formats || this.defaultFormats(info.calendar);
        parseFormats = Array.isArray(parseFormats) ? parseFormats : [parseFormats];

        var length = parseFormats.length;

        for (var idx = 0; idx < length; idx++) {
            date = this.parseExact(parseValue, parseFormats[idx], info);
            if (date) {
                return date;
            }
        }

        return date;
    }
    parseMicrosoftDateFormat(value) {
        if (value && value.indexOf("/D") === 0) {
            var date: any = dateRegExp.exec(value);
            if (date) {
                date = date[1];
                var tzoffset: any = offsetRegExp.exec(date.substring(1));

                date = new Date(parseInt(date, 10));

                if (tzoffset) {
                    tzoffset = this.parseMicrosoftDateOffset(tzoffset[0]);
                    date = this.convertTimeZone(date, date.getTimezoneOffset(), 0);
                    date = this.convertTimeZone(date, 0, -1 * tzoffset);
                }

                return date;
            }
        }
    }

    parseExact(value, format, info) {
        var pattern = this.datePattern(format, info).split(this.EMPTY);

        var state = {
            format: pattern,
            idx: 0,
            value: value,
            valueIdx: 0,
            year: null,
            month: null,
            day: null,
            hours: null,
            minutes: null,
            seconds: null,
            milliseconds: null
        };
        var length = pattern.length;
        var literal = false;

        for (; state.idx < length; state.idx++) {
            var ch = pattern[state.idx];

            if (literal) {
                if (ch === "'") {
                    literal = false;
                }

                this.checkLiteral(state);
            } else {
                if (parsers[ch]) {
                    var invalid = parsers[ch](state, info);
                    if (invalid) {
                        return null;
                    }
                } else if (ch === "'") {
                    literal = true;
                    this.checkLiteral(state);
                } else if (!this.checkLiteral(state)) {
                    return null;
                }
            }
        }

        if (state.valueIdx < value.length) {
            return null;
        }

        return this.createDate(state) || null;
    }
    convertTimeZone(date, fromOffset, toOffset) {
        var fromLocalOffset = date.getTimezoneOffset();

        var offsetDate = new Date(date.getTime() + (fromOffset - toOffset) * 60000);

        var toLocalOffset = offsetDate.getTimezoneOffset();

        return new Date(offsetDate.getTime() + (toLocalOffset - fromLocalOffset) * 60000);
    }

    parseMicrosoftDateOffset(offset) {
        var sign = offset.substr(0, 1) === "-" ? -1 : 1;

        var result = offset.substring(1);
        result = (parseInt(result.substr(0, 2), 10) * 60) + parseInt(result.substring(2), 10);

        return sign * result;
    }
    defaultFormats(calendar) {
        var formats = [];
        var patterns = calendar.patterns;
        var length = this.FORMATS_SEQUENCE.length;

        for (var idx = 0; idx < length; idx++) {
            formats.push(patterns[this.FORMATS_SEQUENCE[idx]]);
        }

        return formats.concat(this.STANDARD_DATE_FORMATS);
    }
    checkLiteral(state) {
        var result = false;
        if (state.value.charAt(state.valueIdx) === state.format[state.idx]) {
            state.valueIdx++;
            result = true;
        }
        return result;
    }

    createDate(state) {
        var year = state.year;
        var month = state.month;
        var day = state.day;
        var hours = state.hours;
        var minutes = state.minutes;
        var seconds = state.seconds;
        var milliseconds = state.milliseconds;
        var pmHour = state.pmHour;
        var UTC = state.UTC;
        var hoursOffset = state.hoursOffset;
        var minutesOffset = state.minutesOffset;
        var hasTime = hours !== null || minutes !== null || seconds || null;
        var date = new Date();
        var result;

        if (year === null && month === null && day === null && hasTime) {
            year = date.getFullYear();
            month = date.getMonth();
            day = date.getDate();
        } else {
            if (year === null) {
                year = date.getFullYear();
            }

            if (day === null) {
                day = 1;
            }
        }

        if (pmHour && hours < 12) {
            hours += 12;
        }

        if (UTC) {
            if (hoursOffset) {
                hours += -hoursOffset;
            }

            if (minutesOffset) {
                minutes += -minutesOffset * (hoursOffset < 0 ? -1 : 1);
            }

            result = new Date(Date.UTC(year, month, day, hours, minutes, seconds, milliseconds));
        } else {
            result = new Date(year, month, day, hours, minutes, seconds, milliseconds);
            this.adjustDST(result, hours);
        }

        if (year < 100) {
            result.setFullYear(year);
        }

        if (result.getDate() !== day && UTC === undefined) {
            return null;
        }

        return result;
    }

    adjustDST(date, hours) {
        if (!hours && date.getHours() === 23) {
            date.setHours(date.getHours() + 2);
        }
    }

    dateFormatNames(options, locale='en-US') {
        var type = options.type;
        var nameType = options.nameType;
        var standAlone = options.standAlone;
        var lower = options.lower;
        var info = this.getLocaleInfo(locale);
        var formatType = standAlone ? "stand-alone" : "format";
        var lowerNameType = (lower ? "lower-" : "") + nameType;
        var formatNames = info.calendar[type][formatType];
        var result = formatNames[lowerNameType];
        if (!result && lower) {
            result = formatNames[lowerNameType] = cloneLower(formatNames[nameType]);
        }
        return result;
    }
    getLocaleInfo(locale) {
        var info;
        if (this.isString(locale)) {
            info = this.localeInfo(locale);
        } else {
            info = locale;
        }
        return info;
    }

    dateFieldName(options, locale) {
        if (locale === void 0) locale = this.DEFAULT_LOCALE;

        var info = this.localeInfo(locale);
        var dateFields = info.calendar.dateFields;

        if (!dateFields) {
            console.log('Data fieldname error!')
        }

        var fieldNameInfo = dateFields[options.type] || {};

        return fieldNameInfo[options.nameType] || fieldNameInfo['wide'];
    }
}