
import { formatRegExp, numberRegExp, numberRegex, timeZoneOffsetRegExp } from './regular-expressions';

var MONTHS_LENGTH = 12;
var DAYS = ["sun", "mon", "tue", "wed", "thu", "fri", "sat"];

function normalize(date, expectedMonth) {
    return (date.getMonth() !== expectedMonth ? lastDayOfMonth(addMonths(date, -1)) : date);
}

function addDays(date, offset) {
    var newDate = cloneDate(date);
    newDate.setDate(newDate.getDate() + offset);
    return adjustDST(newDate, date.getHours());
}

export function cloneDate(date) { 
    return date ? new Date(date.getTime()) : null; 
}

export function isInRange(candidate, min, max) { 
    return (!candidate || !((min && min > candidate) || (max && max < candidate))); 
}

export function isEqual(candidate, expected) {
    if (!candidate && !expected) {
        return true;
    }
    return candidate && expected && candidate.getTime() === expected.getTime();
}

export function dateInRange(candidate, min, max) {
    if (!candidate) {
        return candidate;
    }
    if (min && candidate < min) {
        return cloneDate(min);
    }
    if (max && candidate > max) {
        return cloneDate(max);
    }
    return candidate;
}

export function setTime (origin, candidate) {
    var date = cloneDate(origin);
    date.setHours(candidate.getHours(), candidate.getMinutes(), candidate.getSeconds(), candidate.getMilliseconds());
    return date;
}

export function getDate(date: Date) {
    return createDate(date.getFullYear(), date.getMonth(), date.getDate(), 0, 0, 0, 0);
}

export function createDate(year, month, day, hours, minutes, seconds, milliseconds) {
    if (hours === void 0) { hours = 0; }
    if (minutes === void 0) { minutes = 0; }
    if (seconds === void 0) { seconds = 0; }
    if (milliseconds === void 0) { milliseconds = 0; }
    var date = new Date(year, month, day, hours, minutes, seconds, milliseconds);
    if (year > -1 && year < 100) {
        date.setFullYear(date.getFullYear() - 1900);
    }
    return adjustDST(date, hours);
}

export function adjustDST(date, hour) {
    var newDate = cloneDate(date);
    if (hour === 0 && newDate.getHours() === 23) {
        newDate.setHours(newDate.getHours() + 2);
    }
    return newDate;
}

export function addMonths(date, offset) {
    var newDate = cloneDate(date);
    var diff = (newDate.getMonth() + offset) % MONTHS_LENGTH;
    var expectedMonth = (MONTHS_LENGTH + diff) % MONTHS_LENGTH;
    newDate.setMonth(newDate.getMonth() + offset);
    return normalize(adjustDST(newDate, date.getHours()), expectedMonth);
}

export function lastDayOfMonth (date) {
    var newDate = createDate(date.getFullYear(), date.getMonth() + 1, 1, date.getHours(), date.getMinutes(), date.getSeconds(), date.getMilliseconds());
    return addDays(newDate, -1);
}



export function pad(number, digits, right = null) {
    if (digits === void 0) digits = 2;
    if (right === void 0) right = false;

    var count = digits - String(number).length;
    var result = number;

    if (count > 0) {
        var padString = new Array(count + 1).join("0");
        result = right ? number + padString : padString + number;
    }

    return result;
}
export function  isString(value) {
    return typeof value === "string";
}


export function getLocaleInfo(locale) {
    var info;
    if (isString(locale)) {
        info = this.localeInfo(locale);
    } else {
        info = locale;
    }
    return info;
}

export function lowerArray(arr) {
    var result = [];
    for (var idx = 0; idx < arr.length; idx++) {
        result.push(arr[idx].toLowerCase());
    }
    return result;
}

export function lowerObject(obj) {
    var result = {};
    for (var field in obj) {
        result[field] = obj[field].toLowerCase();
    }
    return result;
}

export function cloneLower(obj) {
    var result = Array.isArray(obj) ? lowerArray(obj) : lowerObject(obj);
    return result;
}

export function dateNameType(formatLength) {
    var nameType;
    if (formatLength <= 3) {
        nameType = "abbreviated";
    } else if (formatLength === 4) {
        nameType = "wide";
    } else if (formatLength === 5) {
        nameType = "narrow";
    } else if (formatLength === 6) {
        nameType = "short";
    }

    return nameType;
}

export function dateFormatNames(locale, options) {
    var type = options.type;
    var nameType = options.nameType;
    var standAlone = options.standAlone;
    var lower = options.lower;
    var info = getLocaleInfo(locale);
    var formatType = standAlone ? "stand-alone" : "format";
    var lowerNameType = (lower ? "lower-" : "") + nameType;
    var formatNames = info.calendar[type][formatType];
    var result = formatNames[lowerNameType];
    if (!result && lower) {
        result = formatNames[lowerNameType] = cloneLower(formatNames[nameType]);
    }
    return result;
}
export function formatNames(locale, type, formatLength, standAlone, lower = null) {
    return dateFormatNames(locale, {
        type: type,
        nameType: dateNameType(formatLength),
        standAlone: standAlone,
        lower: lower
    });
}

export function formatMonth(month, formatLength, info, standAlone) {
    if (formatLength <= 2) {
        return pad(month + 1, formatLength, null);
    }
    return formatNames(info, "months", formatLength, standAlone, null)[month];
}

export function formatString(format) {
    var values = arguments;

    return format.replace(formatRegExp, function (match, index) {
        var value = values[parseInt(index, 10) + 1];

        return value;
    });
}


export function formatTimeZone(date, info, options) {
    var shortHours = options.shortHours;
    var optionalMinutes = options.optionalMinutes;
    var separator = options.separator;
    var localizedName = options.localizedName;
    var zZeroOffset = options.zZeroOffset;
    var offset = date.getTimezoneOffset() / 60;
    if (offset === 0 && zZeroOffset) {
        return "Z";
    }
    var sign = offset <= 0 ? "+" : "-";
    var hoursMinutes = Math.abs(offset).toString().split(".");
    var minutes = hoursMinutes[1] || 0;
    var result = sign + (shortHours ? hoursMinutes[0] : pad(hoursMinutes[0], 2, null));
    if (minutes || !optionalMinutes) {
        result += (separator ? ":" : "") + pad(minutes, 2, null);
    }

    if (localizedName) {
        var localizedFormat = offset === 0 ? info.calendar.gmtZeroFormat : info.calendar.gmtFormat;
        result = formatString(localizedFormat);
    }

    return result;
}

export function firstDay() {
    return DAYS.indexOf("sun");
}


export function formatDayOfWeekIndex(day, formatLength, localeInfo) {
    var firstDayIndex = firstDay();
    var dayIndex;
    if (day < firstDayIndex) {
        dayIndex = 7 - firstDayIndex + day;
    } else {
        dayIndex = day - firstDayIndex;
    }

    return dayIndex + 1;
}

export function formatDayOfWeek(date, formatLength, info, standAlone) {
    var result;
    if (formatLength < 3) {
        result = formatDayOfWeekIndex(date.getDay(), formatLength, info);
    } else {
        result = formatNames(info, "days", formatLength, standAlone, null)[date.getDay()];
    }
    return result;
}

export function formatQuarter(date, formatLength, info, standAlone) {
    var quarter = Math.floor(date.getMonth() / 3);
    if (formatLength < 3) {
        return quarter + 1;
    }

    return formatNames(info, "quarters", formatLength, standAlone, null)[quarter];
}
export function lookAhead(match, state) {
    var format = state.format;
    var idx = state.idx;
    var i = 0;
    while (format[idx] === match) {
        i++;
        idx++;
    }
    if (i > 0) {
        idx -= 1;
    }
    state.idx = idx;
    return i;
}

export function round(value, precision) {
    var result = value;
    var decimals = precision || 0;

    result = result.toString().split('e');
    result = Math.round(Number(result[0] + 'e' + (result[1] ? (Number(result[1]) + decimals) : decimals)));

    result = result.toString().split('e');
    result = Number(result[0] + 'e' + (result[1] ? (Number(result[1]) - decimals) : -decimals));

    return result.toFixed(Math.min(decimals, 20));
}

export function getNumber(size, state) {
    var regex = size ? (numberRegExp[size] || new RegExp('^\\d{1,' + size + '}')) : numberRegex,
        match = state.value.substr(state.valueIdx, size).match(regex);

    if (match) {
        match = match[0];
        state.valueIdx += match.length;
        return parseInt(match, 10);
    }
    return null;
}


export function outOfRange(value, start, end) {
    return !(value >= start && value <= end);
}

export function getIndexByName(names, state, lower) {
    var i = 0,
        length = names.length,
        name, nameLength,
        matchLength = 0,
        matchIdx = 0,
        subValue;

    for (; i < length; i++) {
        name = names[i];
        nameLength = name.length;
        subValue = state.value.substr(state.valueIdx, nameLength);

        if (lower) {
            subValue = subValue.toLowerCase();
        }

        if (subValue === name && nameLength > matchLength) {
            matchLength = nameLength;
            matchIdx = i;
        }
    }

    if (matchLength) {
        state.valueIdx += matchLength;
        return matchIdx + 1;
    }

    return null;
}

export function parseMonth(ch, state, info) {
    var count = lookAhead(ch, state);
    var names = formatNames(info, "months", count, ch === "L", true);

    var month = count < 3 ? getNumber(2, state) : getIndexByName(names, state, true);

    if (month === null || outOfRange(month, 1, 12)) {
        return true;
    }
    state.month = month - 1;
}

export function calendarGmtFormats(calendar) {
    var gmtFormat = calendar.gmtFormat;
    var gmtZeroFormat = calendar.gmtZeroFormat;
    if (!gmtFormat) {
        console.log('CalendarGMTFormats - error');
    }

    return [gmtFormat.replace("{0}", "").toLowerCase(), gmtZeroFormat.replace("{0}", "").toLowerCase()];
}


export function parseTimeZoneOffset(state, info, options) {
    var shortHours = options.shortHours;
    var noSeparator = options.noSeparator;
    var optionalMinutes = options.optionalMinutes;
    var localizedName = options.localizedName;
    var zLiteral = options.zLiteral;
    state.UTC = true;

    if (zLiteral && state.value.charAt(state.valueIdx) === "Z") {
        state.valueIdx++;
        return false;
    }

    if (localizedName && !getIndexByName(calendarGmtFormats(info.calendar), state, true)) {
        return true;
    }

    var matches = timeZoneOffsetRegExp.exec(state.value.substr(state.valueIdx, 6));
    if (!matches) {
        return !localizedName;
    }

    var hoursMatch = matches[1];
    var minutesMatch = matches[3];
    var hoursOffset = parseInt(hoursMatch, 10);
    var separator = matches[2];
    var minutesOffset = parseInt(minutesMatch, 10);

    if (isNaN(hoursOffset) || (!shortHours && hoursMatch.length !== 3) || (!optionalMinutes && isNaN(minutesOffset)) || (noSeparator && separator)) {
        return true;
    }

    if (isNaN(minutesOffset)) {
        minutesOffset = null;
    }

    if (outOfRange(hoursOffset, -12, 13) || (minutesOffset && outOfRange(minutesOffset, 0, 59))) {
        return true;
    }

    state.valueIdx += matches[0].length;
    state.hoursOffset = hoursOffset;
    state.minutesOffset = minutesOffset;
}

export function parseDayOfWeek(ch, state, info) {
    var count = lookAhead(ch, state);
    var names = formatNames(info, "days", count, ch === "c", true);
    var dayOfWeek = count < 3 ? getNumber(1, state) : getIndexByName(names, state, true);
    if ((!dayOfWeek && dayOfWeek !== 0) || outOfRange(dayOfWeek, 1, 7)) {
        return true;
    }
}

export function startOfQuarter() {
    var date = new Date()
    var currentMonth = date.getMonth()
    var month = currentMonth - currentMonth % 3
    date.setMonth(month, 1)
    date.setHours(0, 0, 0, 0)
    return date;
}

export function endOfQuarter() {
    var d = startOfQuarter();
    d.setMonth(d.getMonth() + 2);
    return new Date(d.getFullYear(), d.getMonth() + 1, 0);
}

export function startOfYear() {
    var date = new Date();
    date.setFullYear(date.getFullYear(), 0, 1)
    date.setHours(0, 0, 0, 0)
    return date;
}

export function endOfYear() {
    var date = new Date();
    var year = date.getFullYear()
    date.setFullYear(year + 1, 0, 0)
    date.setHours(0, 0, 0, 0)
    return date;
}

export function startOfLastWeek() {
    var date = new Date();
    date.setDate(date.getDate() - date.getDay() - 7);
    return date;
}

export function endOfLastWeek() {
    var date = new Date();
    date.setDate(date.getDate() - date.getDay() - 1);
    return date;
}

export function startOfLastMonth() {
    var date = new Date();
    date.setDate(1);
    date.setMonth(date.getMonth()-1);
    return date;
}

export function endOfLastMonth() {
    var date = new Date();
    date.setDate(0);
    return date;
}

export function startOfLastQuarter() {
    var date = startOfQuarter();
    date.setMonth(date.getMonth() - 3);
    return date;
}

export function endOfLastQuarter() {
    var date = endOfQuarter();
    date.setMonth(date.getMonth() - 3);
    date.setDate(date.getDate() -1);
    return date;
}

export function startOfLastYear() {
    var date = new Date();
    var year = date.getFullYear()
    date.setFullYear(year - 1, 0, 1)
    date.setHours(0, 0, 0, 0)
    return date;
}

export function endOfLastYear() {
    var date = new Date();
    var year = date.getFullYear()
    date.setFullYear(year - 1, 11, 31)
    date.setHours(0, 0, 0, 0)
    return date;
}
