import {
  Component,
  forwardRef,
  Input,
  Output,
  EventEmitter,
  ViewChild,
  ElementRef,
  ContentChildren,
  TemplateRef,
  QueryList
} from "@angular/core";
import { AbstractValueAccessor } from "../../baseclasses/abstract-value.accessor";
import { NG_VALUE_ACCESSOR } from "@angular/forms";
import { TemplateParserComponent } from "../../baseclasses/template-parser.component-base";
import { CbxTemplateDirective } from "../../directives/cbx-template.directive";

const CBX_TOGGLE_CONTROL_VALUE_ACCESSOR: any = {
  provide: NG_VALUE_ACCESSOR,
  useExisting: forwardRef(() => CbxToggleComponent),
  multi: true
};

@Component({
  selector: "cbx-toggle",
  templateUrl: "./cbx-toggle.component.html",
  styleUrls: ["./cbx-toggle.component.less"],
  providers: [CBX_TOGGLE_CONTROL_VALUE_ACCESSOR],
  host:{
    '[class.cbx]':'true',
    '[class.cbx-state-disabled]':'disabled'
  }
})
export class CbxToggleComponent extends AbstractValueAccessor {
  @Input() label:string = "";
  @Input()
  allowNull: boolean = false;
  @Input() required: boolean = false;
  @Input()
  className: string;
  @Input() disabled:boolean = false;

  _defaultOption: number = 0;
  @Input("defaultOption")
  set defaultOption(value: number) {
    this._defaultOption = value;
  }
  get defaultOption() {
    return this._defaultOption;
  }

  @Input() strTooltip:string = "";

  defaultOptionLength: number = 20;

  @Output()
  clicked: EventEmitter<boolean> = new EventEmitter();


  //values and labels --------- start
  //one defaults to true, two defaults to false
  //https://corebridge.atlassian.net/wiki/spaces/ENDOR/pages/806584332/Custom+Field+Visual+Components
  @Input() optionOneValue:any =  true;
  @Input() optionTwoValue:any =  false;

  private _optionOneLabel:string = null;
  @Input() get optionOneLabel():string{
    return this._optionOneLabel || 'Yes';
  }
  set optionOneLabel(val:string){
    this._optionOneLabel = val;
  }

  private _optionTwoLabel:string = null;
  @Input() get optionTwoLabel():string{
    return this._optionTwoLabel || 'No';
  }
  set optionTwoLabel(val:string){
    this._optionTwoLabel = val;
  }
  //values and labels --------- end

  public get on():boolean{
    if(this.value === this.optionOneValue){
      return true;
    }else if(this.value === this.optionTwoValue){
      return false;
    }else if(this.allowNull){
      return null;
    }else if(!this.allowNull){
      return false;
    }
  }

  userClick: boolean = false;
  @ViewChild("button")
  button: ElementRef;

  toggleClick(ev:MouseEvent) {
    //console.log(ev);

    if(this.disabled){
      return;
    }
        
    if (this.allowNull) {//if tri state
      if (ev.srcElement != this.button.nativeElement) {
        return;
      }
      var hitIndex = this.getHitIndex(ev);

      if(hitIndex===null){//if enter key was used
        if(this.value == this.optionOneValue || this.value == this.optionTwoValue){
          this.value = null;
        }else{
          if(this._previousValue==this.optionOneValue){
            this.value = this.optionTwoValue;
          }else{
            this.value = this.optionOneValue;
          }
        }
        this.updateValue();

      }else{//if clik was used
        switch (hitIndex) {
          case 0://hit on left side
            this.value = this.optionTwoValue;
            break;
          case 1://hit on center
            this.value = null;//null
            break;
          case 2://hit on right side
            this.value = this.optionOneValue;
            break;
        }
      }

    } else {//if dual state
      this.value = this.value===this.optionOneValue? this.optionTwoValue:this.optionOneValue;
    }
  }

  updateValue() {
    //var selectedOption = this.getSelectedOption();
    this.clicked.emit(this.value);
  }

  /**
   * 0=left, 1=center, 2=right
   * @param event `MouseEvent`
   */
  getHitIndex(event) {
    let isEnterKey = event.screenX==0 && event.screenY==0;
    if(isEnterKey){
      return null;
    }

    var numOfSegment = 2;
    if (this.allowNull) {
      numOfSegment++;
    }
    var segmentWidth = event.srcElement.offsetWidth / numOfSegment;
    var relativeHit = event.offsetX;

    for (var x = 1; x <= numOfSegment; x++) {
      if (relativeHit < x * segmentWidth) {
        return x - 1;
      }
    }
    return 0;
  }

  public writeValue(value: boolean) {
    this._value = value;
    super.writeValue(value);
  }

  optionClick(ev:Event, val:any){
    ev && ev.stopPropagation();
    if(this.disabled){
      return;
    }
    this.value = val;
    this.updateValue();
  }

  public templateDictionary:{[strname:string]:TemplateRef<any>} = {}
  @ContentChildren(CbxTemplateDirective) templateList:QueryList<CbxTemplateDirective>;

  ngAfterContentInit(){
      if(this.templateList){
          this.templateList.forEach(tpl => this.templateDictionary[tpl.name]=tpl.templateRef);
      }
  }  
}
