import { Directive, ElementRef, EventEmitter, Output, Input, SimpleChange, SimpleChanges, LOCALE_ID, Inject, ChangeDetectorRef, OnInit, OnDestroy } from "@angular/core";
import { CbxInputWrapperComponent } from "../components/cbx-input-wrapper/cbx-input-wrapper.component";
import { NgControl } from "@angular/forms";
import { NumberUtilitiesService } from "../services/number-utilities.service";
import { CurrencyPipe } from "@angular/common";
import { Subscription } from "rxjs";

const blankvals:any[] = [null, undefined, ''];

enum cbxInputDisplayFormats{
  General = "General",
  FixedDecimalPlaces = "Fixed",
  Currency = "Currency",
  Accounting = "Accounting"
}

@Directive({
  selector: "input[cbxInput]",
  host: {
    "[class.cbx]": "true",
    "[class.has-value]": "wrapper.hasValue",
    "[class.hide-spinner]": "!cbxShowSpinner",
    "(input)": "onInput($event)",
    "(keypress)": "onKeypress($event)",
    "(mouseenter)": "toggleHover(true)",
    "(mouseleave)": "toggleHover(false)",
    "(focus)": "onInputFocus()",
    "(blur)": "onInputBlur()",
  }
})
export class CbxInputDirective implements OnInit, OnDestroy {

  @Input() cbxShowSpinner:boolean = true;
  @Input() cbxAllowDecimals:boolean = true;
  @Input() cbxDecimalsPlaces:number = null;
  /**
   *used only to store original input.type in case we want special formatting
   *
   *note: do not user <input [type]="thetypeyouwant">, use <input type="thetypeyouwant">,
   *for some unkown reason angular tracks the model as a string if we use <input [type]="'number'">
   */
  @Input('type') inputtype:string = "";

  @Input() cbxInputDisplayFormat:any = null;//using any so we can either bind a string or an enum value

  @Output() cbxInputError:EventEmitter<any> = new EventEmitter();

  input: HTMLInputElement;

  private currPipe:CurrencyPipe;
  private subs = new Subscription();

  constructor(elemRef: ElementRef, public wrapper: CbxInputWrapperComponent,
    private ngControl:NgControl,
    private numUtilSvc: NumberUtilitiesService,
    @Inject(LOCALE_ID) private locale: string,
    private cdRef:ChangeDetectorRef){
    this.input = elemRef.nativeElement;
    this.currPipe = new CurrencyPipe(locale);
  }

  ngOnInit(): void {
    this.subs.add(this.ngControl.valueChanges.subscribe(val => {
      this.attachFormat();
    }));
  }

  ngOnDestroy(): void {
    this.subs.unsubscribe();
  }

  ngOnChanges(changes:SimpleChanges){
    if(changes.cbxDecimalsPlaces){
      this.validateDecimals();
    }
    this.attachFormat();
  }

  ngDoCheck() {
    this.checkHasValue();
    this.checkIsNegative();
    this.wrapper.displayFormat = (this.cbxInputDisplayFormat || "none").toLowerCase();
  }

  checkHasValue() {
    this.wrapper.hasValue = (this.input.value || "").length > 0;
  }

  checkIsNegative(){
    if(isNaN(this.ngControl.value)){
      this.wrapper.isNegative = false;
      return;
    }

    this.wrapper.isNegative = Number(this.ngControl.value) < 0;
  }

  onInput(ev:Event){
    this.checkHasValue();
    this.validateDecimals();
  }

  onKeypress(ev:KeyboardEvent){
    //do something
    // var charCode = (ev.which) ? ev.which : ev.keyCode
    // return !(charCode > 31 && (charCode < 48 || charCode > 57) && charCode != 46);
  }

  /**
   * firefox fires a blur event when the input.type is changed
   *
   * this is a bug in firefox itself https://bugzilla.mozilla.org/show_bug.cgi?id=981248
   */
  private inputFocusBusy = false;

  onInputFocus(){
    this.inputFocusBusy = true;

    this.toggleFocus(true);
    this.detachFormat();

    setTimeout(() => {
      this.inputFocusBusy = false;
    }, 100);

  }

  onInputBlur() {
    if(this.inputFocusBusy){
      return;
    }

    if(blankvals.includes(this.input.value) && this.input.validity.badInput && this.input.type=="number"){
      this.cbxInputError.emit(this.getCurrentInputValidity());
      this.input.value = null;
    }

    this.toggleFocus(false);

    this.attachFormat();
  }

  toggleHover(hasHover: boolean) {
    this.wrapper.hasHover = hasHover;
  }

  toggleFocus(hasFocus: boolean) {
    this.wrapper.hasFocus = hasFocus;
  }

  /**
   * store the actual current value of this.input.validity
   * for some reason ...syntax is not working
   */
  private getCurrentInputValidity():ValidityState{
    return {
      badInput: this.input.validity.badInput,
      customError: this.input.validity.customError,
      patternMismatch: this.input.validity.patternMismatch,
      rangeOverflow: this.input.validity.rangeOverflow,
      rangeUnderflow: this.input.validity.rangeUnderflow,
      stepMismatch: this.input.validity.stepMismatch,
      tooLong: this.input.validity.tooLong,
      tooShort: this.input.validity.tooShort,
      typeMismatch: this.input.validity.typeMismatch,
      valid: this.input.validity.valid,
      valueMissing: this.input.validity.valueMissing
    }
  }

  validateDecimals(){
    if(this.inputtype!=='number'){
      return;
    }

    if(this.cbxAllowDecimals && (this.cbxDecimalsPlaces === null || this.wrapper.hasFocus)){
      return;
    }

    let hasValue = this.numUtilSvc.hasValue(this.ngControl.value);
    if(!hasValue){
      return;
    }

    let decimals = this.numUtilSvc.countDecimals(this.ngControl.value);

    if(hasValue && decimals>1  && !this.cbxAllowDecimals){
      this.setModelAndInput(this.numUtilSvc.trimDecimals(this.ngControl.value, 0));
      return;
    }

    if(hasValue && decimals>this.cbxDecimalsPlaces){
      this.setModelAndInput(this.numUtilSvc.trimDecimals(this.ngControl.value, this.cbxDecimalsPlaces));
      return;
    }

  }

  setModelAndInput(val:any){
    this.ngControl.control.setValue(val);
  }

  attachFormat() {
    if (this.wrapper.hasFocus || this.cbxInputDisplayFormat === null) {
      return; //do nothing if user is currently using the input or cbxInputDisplayFormat is null
    }

    this.input.type = "text"
    //do formatting here below
    switch (this.cbxInputDisplayFormat) {

      case cbxInputDisplayFormats.General: {
        this.input.value = this.ngControl.value;
        break;
      }
      case cbxInputDisplayFormats.FixedDecimalPlaces: {
        // add zeroes on vacant decimal palces depend on decimal places limit
        if (this.numUtilSvc.hasValue(this.ngControl.value)) {
          let parsedValue = parseFloat(this.ngControl.value);
          this.input.value = parsedValue.toFixed(this.cbxDecimalsPlaces || 0);
        }
        break;
      }
      case cbxInputDisplayFormats.Currency: {
        var digits:string = '1.'+this.cbxDecimalsPlaces.toString()+'-'+this.cbxDecimalsPlaces.toString();
        this.input.value = this.currPipe.transform(this.ngControl.value,'USD','symbol',digits);
        break;
      }
      case cbxInputDisplayFormats.Accounting: {
        //https://stackoverflow.com/questions/24758817/javascript-number-tolocalestring-with-4-digits-after-separator

        if (this.numUtilSvc.hasValue(this.ngControl.value)) {
          let decimalPlaces = this.cbxDecimalsPlaces || 0;
          let localeString = parseFloat(Math.abs(this.ngControl.value).toString())
                              .toLocaleString(this.locale, { minimumFractionDigits: decimalPlaces, maximumFractionDigits: decimalPlaces });

          if (Number(this.ngControl.value) < 0)
            this.input.value = `(${localeString})`;
          else
            this.input.value = localeString;
        }
        break;
      }
    }
  }

  detachFormat() {
    this.input.type = this.inputtype;

    //do unformatting below
    //your code here
    if (this.numUtilSvc.hasValue(this.ngControl.value)){
      this.input.value = this.ngControl.value;
    }
  }

}
