import { Directive, Input, TemplateRef, ApplicationRef, ComponentFactoryResolver, Injector, EmbeddedViewRef, ComponentRef, ElementRef, HostListener } from '@angular/core';
import { CbxTooltipContainerComponent } from '../components/cbx-tooltip-container/cbx-tooltip-container.component';
import Popper from 'popper.js';
import { GlobalEventsService } from '../services/global-events.service';
import { Subscription } from 'rxjs';

export interface cbxToolTipVisbilityConfig{
  showOnHover?:boolean;
  hideOnHoverout?:boolean;
  showOnClick?:boolean;
  hideOnClickout?:boolean;
  showOnFocus?:boolean;
  hideOnBlur?:boolean;
}

/** https://hackernoon.com/angular-pro-tip-how-to-dynamically-create-components-in-body-ba200cc289e6 */
@Directive({
  selector: '[cbxTooltip]'
})
export class CbxTooltipDirective {

  constructor(private appRef:ApplicationRef,
              private componentFactoryResolver:ComponentFactoryResolver,
              private injector:Injector,
              private elemRef:ElementRef,
              private globalEvSvc:GlobalEventsService
              ) { }

  @Input('cbxTooltip') set _tooltipContent(val:string|TemplateRef<any>){
    this.tooltipContent = val;
    this.resolveInputs();
  }

  @Input('cbxTooltipPlacement') set _tooltipPlacement(val:string){
    this.tooltipPlacement = val;
    this.resolveInputs();
  }

  @Input('cbxTooltipClassType') set _tooltipClassType(val:string){
    this.tooltipClassType = val;
    this.resolveInputs();
  }

  @Input('cbxTooltipFontSize') set _tooltipFontSize(val:string){
    this.tooltipFontSize = val;
    this.resolveInputs();
  }


  @Input('cbxTooltipDisabled') set _tooltipDisabled(val:any){
    this.tooltipDisabled = !!val;//double negation to ensure boolean value
    this.resolveInputs();
  }

  @Input() cbxToolTipVisbility:cbxToolTipVisbilityConfig = {
    showOnHover: true,
    hideOnHoverout: true
  }

  tooltipContent:string|TemplateRef<any> = null;
  tooltipPlacement:string = 'bottom';//https://popper.js.org/popper-documentation.html#Popper.Defaults.placement
  private initialized:boolean = false;

  public tooltipClassType:string = 'cbx-tooltip-default';
  public tooltipFontSize:string = 'cbx-tooltip-text-medium';
  public tooltipDisabled:boolean = false;
  

  private containerRef:ComponentRef<CbxTooltipContainerComponent> = null;
  private popper:Popper = null;
  private isHostHovered:boolean = false;
  private subs = new Subscription();

  ngOnInit(){
    this.initialized = true;
    this.createPopperContainer();
    this.createPopper();
    this.subs.add(this.globalEvSvc.FocusMapStream.subscribe(focusMap=>this.onClickOut(focusMap)));
  }

  ngOnDestroy(){
    this.containerRef.destroy();
    this.popper.destroy();
    this.subs.unsubscribe();
  }

  createPopperContainer(){
    let componentFactory = this.componentFactoryResolver.resolveComponentFactory(CbxTooltipContainerComponent);
    this.containerRef = componentFactory.create(this.injector);
    
    this.appRef.attachView(this.containerRef.hostView);
    this.resolveInputs();

    const domElem = (this.containerRef.hostView as EmbeddedViewRef<any>).rootNodes[0] as HTMLElement;

    document.body.appendChild(domElem);
  }

  createPopper(){
    this.popper = new Popper(this.elemRef.nativeElement, (this.containerRef.hostView as EmbeddedViewRef<any>).rootNodes[0] as HTMLElement);
    this.resolveInputs();
    setTimeout(()=>(this.popper && this.popper.update()), 150);
  }

  resolveInputs(){
    if(this.popper){
      this.popper.options.placement = this.tooltipPlacement.replace("-start","").replace("-end","") as any;
      this.popper.update();
      this.popper.scheduleUpdate();
    }

    if(this.containerRef){
      this.containerRef.instance.tooltipClassType = this.tooltipClassType;
      this.containerRef.instance.tooltipFontSize = this.tooltipFontSize;
      this.containerRef.instance.tooltipString = null;
      this.containerRef.instance.tooltipTemplate = null;
      this.containerRef.instance.placement = this.tooltipPlacement;
      this.containerRef.instance.tooltipDisabled = this.tooltipDisabled;
      
      if((typeof this.tooltipContent)==="string"){
        this.containerRef.instance.tooltipString = this.tooltipContent as string;
      }else{
        this.containerRef.instance.tooltipTemplate = this.tooltipContent as TemplateRef<any>;
      }

      this.containerRef.changeDetectorRef.detectChanges();

      this.togglePopperEventListeners();
    }
  }

  show(){
    this.togglePopperVisibility(true);
  }

  hide(){
    this.togglePopperVisibility(false);
  }

  togglePopperVisibility(isShown:boolean){
    if(this.containerRef){
      this.containerRef.instance.isShown = isShown;
      this.containerRef.changeDetectorRef.detectChanges(); 
    }

    this.togglePopperEventListeners();

    if(this.popper){
      this.popper.scheduleUpdate();
    }
  }

  togglePopperEventListeners(){
    if(this.popper && this.containerRef){
      if(this._tooltipDisabled || !this.containerRef.instance.isShown){
        this.popper.disableEventListeners();
      }else{
        this.popper.enableEventListeners();
      }
    }
  }

  @HostListener('mouseenter', ['$event']) onHover(ev:MouseEvent){
    this.isHostHovered = true;

    if(this.cbxToolTipVisbility.showOnHover){
      setTimeout(() => {
        this.isHostHovered && this.show();//after a brief timeout, show the popup if host is still hovered
      }, 500);
    }
  }

  @HostListener('mouseleave', ['$event']) onHoverout(ev:MouseEvent){
    this.isHostHovered = false;

    if(this.cbxToolTipVisbility.hideOnHoverout){
      setTimeout(() => {
        !this.isHostHovered && this.hide();//after a brief timeout, hide the popup if host is no longer hovered
      }, 350);
    }
  }

  @HostListener('click', ['$event']) onClick(ev:MouseEvent){
    if(this.cbxToolTipVisbility.showOnClick){
      this.show();
    }
  }

  onClickOut(focusMap:{click: Event, map: Map<Node, boolean>}){
    const ev = focusMap.click;
    if(ev.target == this.elemRef.nativeElement){
      return;
    }
    if(this.containerRef && ev.target == this.containerRef.instance.hostElem){
      return;
    }
    if(focusMap.map.has(this.elemRef.nativeElement)){
      return;
    }
    if(this.containerRef && focusMap.map.has((this.containerRef.instance.hostElem as HTMLElement))){
      return;
    }

    //if clicked out
    if(this.cbxToolTipVisbility.hideOnClickout){
      this.hide();
    }
  }

  @HostListener('focus', ['$event']) onFocus(ev:MouseEvent){
    if(this.cbxToolTipVisbility.showOnFocus){
      this.show();
    }
  }

  @HostListener('blur', ['$event']) onBlur(ev:MouseEvent){
    if(this.cbxToolTipVisbility.hideOnBlur){
      this.hide();
    }
  }



}
